<!DOCTYPE html>
<html>
  <head>
    <title>Treetable Prototype</title>
    <meta charset="utf-8" />

    <style media="screen">
      /* Add space between Vega-Embed links  */
      .vega-actions a {
        margin-right: 5px;
      }
    </style>
  </head>
  <body>
    <h1>Treetable Prototype</h1>
    <!-- Container for the visualization -->
    <div id="treetable"></div>
  </body>
</html>